import json
import logging
import os
import time
import warnings
from pathlib import Path
from typing import Annotated, Callable, Optional

import numpy as np
import rasterio
import xarray as xr
from fastapi import FastAPI, HTTPException, Query, Response
from fastapi.middleware.cors import CORSMiddleware
from pyproj import CRS, Transformer
from rio_tiler.io.xarray import XarrayReader
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from starlette_cramjam.middleware import CompressionMiddleware
from titiler.core.algorithm import algorithms as default_algorithms
from titiler.core.dependencies import DefaultDependency
from titiler.core.errors import DEFAULT_STATUS_CODES, add_exception_handlers
from titiler.core.factory import TilerFactory

from .config import settings
from .readers import CustomReader
from .titiler import Amplitude, JSONResponse, Phase, RasterGroup, Rewrap, Shift
from .utils import desensitize_mpl_case, generate_colorbar

logger = logging.getLogger("bowser")
warnings.filterwarnings(
    "ignore", category=RuntimeWarning, message="invalid value encountered in cast"
)

t0 = time.time()
# Set up logging
h = logging.StreamHandler()
h.setLevel(settings.LOG_LEVEL)
h.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%dT:%H:%M:%S",
    )
)
logger.addHandler(h)

template_dir = Path(__file__).parent / "templates"
templates = Jinja2Templates(directory=template_dir)
desensitize_mpl_case()

app = FastAPI(
    title="Bowser",
    openapi_url="/api",
    docs_url="/api.html",
)


def load_data_sources():
    """Load data sources and determine which mode to use (MD or COG)."""
    # Check for xarray stack file first (MD mode)
    if settings.BOWSER_STACK_DATA_FILE:
        stack_file = os.environ["BOWSER_STACK_DATA_FILE"]
        logger.info(f"Loading xarray dataset from {stack_file} (MD mode)")
        ds = (
            xr.open_zarr(stack_file)
            if stack_file.endswith(".zarr")
            else xr.open_dataset(stack_file)
        )
        crs = CRS.from_wkt(ds.spatial_ref.crs_wkt)
        ds.rio.write_crs(crs, inplace=True)
        transformer = Transformer.from_crs(4326, ds.rio.crs, always_xy=True)
        return "md", ds, None, transformer

    # Check for RasterGroup JSON config (COG mode)
    elif Path(settings.BOWSER_DATASET_CONFIG_FILE).exists():
        logger.info(
            f"Loading RasterGroups from {settings.BOWSER_DATASET_CONFIG_FILE} (COG)"
        )
        data_js_list = json.loads(Path(settings.BOWSER_DATASET_CONFIG_FILE).read_text())
        raster_groups = {}
        for d in data_js_list:
            rg = RasterGroup.model_validate(d)
            raster_groups[rg.name] = rg
        logger.info(
            (
                f"Found {len(raster_groups)} RasterGroup configs:"
                " {list(raster_groups.keys())}"
            )
        )
        return "cog", None, raster_groups, None

    else:
        raise ValueError(
            "No data files specified - need either stack file or JSON config"
        )


# Global data sources - load once at startup
DATA_MODE, XARRAY_DATASET, RASTER_GROUPS, transformer_from_lonlat = load_data_sources()
print(f"Data loading complete in {time.time() - t0:.1f} sec. Mode: {DATA_MODE}")


def create_xarray_dataset_info(ds: xr.Dataset) -> dict:
    """Create dataset info structure from Xarray Dataset."""
    # Get bounds from the dataset
    bounds = ds.rio.bounds()
    if ds.rio.crs != rasterio.crs.CRS.from_epsg(4326):
        mdim_vars = [v for v in ds.data_vars.values() if v.ndim >= 2]
        # Get just one layer to reproject, and subsample for quick reprojecting
        da = mdim_vars[0][..., ::10, ::10]
        if da.ndim == 3:
            da = da[0]
        latlon_bounds = da.rio.reproject("epsg:4326", subsample=10).rio.bounds()
    else:
        latlon_bounds = bounds
    logger.info(
        f"latlon_bounds: {latlon_bounds}, bounds: {bounds}, ds.rio.crs: {ds.rio.crs}"
    )

    # Get time values formatted for frontend
    time_values = ds.time.dt.strftime("%Y-%m-%d").values.tolist()

    # Create info for each variable that has spatial dimensions
    dataset_info = {}

    skip_spatial_reference = os.getenv("BOWSER_SPATIAL_REFERENCE_DISP") == "NO"
    for var_name, var in ds.data_vars.items():
        if "x" in var.dims and "y" in var.dims:
            use_moving_reference = (
                "displacement" in str(var_name).lower()
                and "short_wave" not in str(var_name).lower()
                and not skip_spatial_reference
            )
            dataset_info[var_name] = {
                "name": var_name,
                "file_list": [
                    f"variable:{var_name}:time:{i}" for i in range(len(ds.time))
                ],
                "mask_file_list": [],
                "mask_min_value": 0.1,
                "nodata": None,
                "uses_spatial_ref": use_moving_reference,
                "algorithm": "shift" if use_moving_reference else None,
                "bounds": list(bounds),
                "latlon_bounds": list(latlon_bounds),
                "x_values": time_values,
            }

    return dataset_info


def create_rastergroup_dataset_info(raster_groups: dict) -> dict:
    """Create dataset info structure from RasterGroups."""
    dataset_info = {}
    for name, rg in raster_groups.items():
        dataset_info[name] = {
            "name": rg.name,
            "file_list": [str(f) for f in rg.file_list],
            "mask_file_list": [str(f) for f in rg.mask_file_list],
            "mask_min_value": rg.mask_min_value,
            "nodata": rg.nodata,
            "uses_spatial_ref": rg.uses_spatial_ref,
            "algorithm": rg.algorithm,
            "bounds": list(rg.bounds),
            "latlon_bounds": list(rg.latlon_bounds),
            "x_values": rg.x_values,
        }
    return dataset_info


@app.get("/datasets")
async def datasets():
    """Return the JSON describing all available datasets."""
    if DATA_MODE == "md":
        return create_xarray_dataset_info(XARRAY_DATASET)
    else:  # cog mode
        return create_rastergroup_dataset_info(RASTER_GROUPS)


@app.get("/mode")
async def get_mode():
    """Return the current data mode (md or cog) for frontend routing."""
    return {"mode": DATA_MODE}


@app.get("/variables")
async def get_variables():
    """Get available variables (only for MD mode)."""
    if DATA_MODE != "md":
        return {"variables": {}}

    variables = {}
    for var_name, var in XARRAY_DATASET.data_vars.items():
        if "x" in var.dims and "y" in var.dims:
            variables[var_name] = {
                "dimensions": list(var.dims),
                "shape": list(var.shape),
                "dtype": str(var.dtype),
                "attrs": dict(var.attrs),
            }
    return {"variables": variables}


async def _get_point_values(dataset_name: str, lon: float, lat: float) -> np.ndarray:
    """Get point values for a dataset at lon/lat."""
    if DATA_MODE == "md":
        if dataset_name not in XARRAY_DATASET.data_vars:
            raise HTTPException(
                status_code=404, detail=f"Variable {dataset_name} not found"
            )
        x, y = transformer_from_lonlat.transform(lon, lat)
        point_data = XARRAY_DATASET[dataset_name].sel(x=x, y=y, method="nearest")
        return np.atleast_1d(point_data.values)
    else:  # cog mode
        if dataset_name not in RASTER_GROUPS:
            raise HTTPException(
                status_code=404, detail=f"Dataset {dataset_name} not found"
            )
        return np.atleast_1d(RASTER_GROUPS[dataset_name]._reader.read_lonlat(lon, lat))


@app.get(
    "/point",
    response_class=JSONResponse,
    responses={200: {"description": "Return the list of values for a point"}},
)
async def point(
    dataset_name: str,
    lon: Annotated[float, Query(..., title="Longitude", ge=-180, le=180)],
    lat: Annotated[float, Query(..., title="Latitude", ge=-90, le=90)],
):
    """Fetch list of time series values for a point."""
    values = await _get_point_values(dataset_name, lon, lat)
    return JSONResponse(values.tolist())


@app.get(
    "/chart_point",
    response_class=JSONResponse,
    responses={200: {"description": "Return the Chart.js-formatted data for a point"}},
)
async def chart_point(
    dataset_name: str,
    lon: Annotated[float, Query(..., title="Longitude", ge=-180, le=180)],
    lat: Annotated[float, Query(..., title="Latitude", ge=-90, le=90)],
    ref_lon: Annotated[
        float | None, Query(..., title="Reference longitude", ge=-180, le=180)
    ] = None,
    ref_lat: Annotated[
        float | None, Query(..., title="Reference latitude", ge=-90, le=90)
    ] = None,
):
    """Fetch the Chart.js time series values for a point (relative to a reference)."""
    # Get time values based on data mode
    if DATA_MODE == "md":
        if dataset_name not in XARRAY_DATASET.data_vars:
            raise HTTPException(
                status_code=404, detail=f"Variable {dataset_name} not found"
            )
        da = XARRAY_DATASET[dataset_name]
        x_values = da.time.dt.strftime("%Y-%m-%d").values.tolist()
    else:  # cog mode
        if dataset_name not in RASTER_GROUPS:
            raise HTTPException(
                status_code=404, detail=f"Dataset {dataset_name} not found"
            )
        rg = RASTER_GROUPS[dataset_name]
        x_values = rg.x_values

    def values_to_chart_data(values):
        """Convert values to chart data."""
        return [
            {"x": x, "y": float(y)}
            for x, y in zip(x_values, np.atleast_1d(values).tolist())
            if not np.isnan(y)
        ]

    # Get values at the point
    values = await _get_point_values(dataset_name, lon, lat)

    # Apply reference subtraction if provided
    if ref_lon is not None and ref_lat is not None:
        ref_values = await _get_point_values(dataset_name, ref_lon, ref_lat)
        values = values - ref_values

    dataset_item = values_to_chart_data(values.tolist())

    return JSONResponse({"datasets": [{"data": dataset_item}], "labels": x_values})


@app.get("/colorbar/{cmap_name}")
async def get_colorbar(cmap_name: str):
    """Get a WEBP image of a matplotlib colorbar."""
    try:
        colorbar_bytes = generate_colorbar(cmap_name)
        return Response(content=colorbar_bytes, media_type="image/webp")
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Set up CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(
    CompressionMiddleware,
    minimum_size=0,
    exclude_mediatype={
        "image/jpeg",
        "image/jpg",
        "image/png",
        "image/jp2",
        "image/webp",
    },
    compression_level=6,
)

# Set up algorithms
algorithms = default_algorithms.register(
    {"phase": Phase, "amplitude": Amplitude, "shift": Shift, "rewrap": Rewrap}
)
PostProcessParams: Callable = algorithms.dependency


# Set up routing based on data mode
# if DATA_MODE == "cog":
# COG mode: use RasterGroup approach
def InputDependency(
    url: Annotated[str, Query(description="Dataset URL")],
    mask: Annotated[str | None, Query(description="Mask URL")] = None,
    mask_min_value: Annotated[float, Query(description="Mask Minimum Value")] = 0.1,
) -> dict:
    """Create dataset path from args."""
    return {
        "data": url,
        "mask": mask,
        "mask_min_value": mask_min_value,
    }


cog_endpoints = TilerFactory(
    router_prefix="/cog",
    reader=CustomReader,
    path_dependency=InputDependency,
    process_dependency=PostProcessParams,
)
app.include_router(
    cog_endpoints.router, prefix="/cog", tags=["Cloud Optimized GeoTIFF"]
)
logger.info("Configured COG endpoints at /cog/*")


# MD mode: use xarray approach
def XarrayPathDependency(
    variable: str = Query(..., description="Variable name"),
    time_idx: Optional[int] = Query(None, description="Time index"),
    mask_variable: Optional[str] = Query(None, description="Mask variable"),
    mask_min_value: float = Query(0.1, description="Mask minimum value"),
) -> xr.DataArray:
    """Create a DataArray from query parameters."""
    da = XARRAY_DATASET[variable]
    skip_recommended_mask = os.getenv("BOWSER_USE_RECOMMENDED_MASK") == "NO"
    if mask_variable is not None:
        mask_da = XARRAY_DATASET[mask_variable]
    elif variable == "displacement" and (
        "recommended_mask" in XARRAY_DATASET.data_vars and not skip_recommended_mask
    ):
        mask_da = XARRAY_DATASET["recommended_mask"]
    else:
        mask_da = None

    if time_idx is None or "time" not in da.dims:
        if mask_da is not None:
            da = da.where(mask_da > mask_min_value)
        return da
    da = da.sel(time=XARRAY_DATASET.time[time_idx])
    if mask_da is not None:
        mask_da = mask_da.sel(time=XARRAY_DATASET.time[time_idx])
        return da.where(mask_da > mask_min_value)
    return da


md_endpoints = TilerFactory(
    router_prefix="/md",
    reader=XarrayReader,
    path_dependency=XarrayPathDependency,
    reader_dependency=DefaultDependency,
    process_dependency=PostProcessParams,
)
app.include_router(md_endpoints.router, prefix="/md", tags=["Xarray Multi Dimensional"])
logger.info("Configured MD endpoints at /md/*")

# Add exception handlers
add_exception_handlers(app, DEFAULT_STATUS_CODES)

# Serve static files
dist_path = Path(__file__).parent / "dist"
app.mount("/", StaticFiles(directory=dist_path, html=True))
print(f"Setup complete: time to load datasets: {time.time() - t0:.1f} sec.")
logger.info(f"Bowser started in {DATA_MODE.upper()} mode")



from pathlib import Path

from .titiler import Algorithm

CORE_DATASETS = [
    "displacement",
    "short_wavelength_displacement",
    "recommended_mask",
    "connected_component_labels",
    "temporal_coherence",
    "estimated_phase_quality",
    "persistent_scatterer_mask",
    "shp_counts",
    "water_mask",
    "phase_similarity",
    "timeseries_inversion_residuals",
]
CORRECTION_DATASETS = [
    "corrections/ionospheric_delay",
    "corrections/perpendicular_baseline",
    "corrections/solid_earth_tide",
]


def get_disp_s1_outputs(disp_s1_dir: Path | str):
    def _glob(pattern: str, subdir: str) -> list[str]:
        return [str(p) for p in sorted((Path(disp_s1_dir) / subdir).glob(pattern))]

    return [
        {
            "name": "Displacement",
            "file_list": _glob("*.vrt", subdir="displacement"),
            "uses_spatial_ref": True,
            "algorithm": Algorithm.SHIFT.value,
            "mask_file_list": _glob("*.vrt", subdir="connected_component_labels"),
        },
        {
            "name": "Short Wavelength Displacement",
            "file_list": _glob("*.vrt", subdir="short_wavelength_displacement"),
        },
        {
            "name": "Connected Component Labels",
            "file_list": _glob("*.vrt", subdir="connected_component_labels"),
        },
        {
            "name": "Re-wrapped phase",
            "file_list": _glob("*.vrt", subdir="displacement"),
            "algorithm": Algorithm.REWRAP.value,
        },
        {
            "name": "Persistent Scatterer Mask",
            "file_list": _glob("*.vrt", subdir="persistent_scatterer_mask"),
        },
        {
            "name": "Temporal Coherence",
            "file_list": _glob("*.vrt", subdir="temporal_coherence"),
        },
        {
            "name": "Phase Similarity",
            "file_list": _glob("*.vrt", subdir="phase_similarity"),
        },
        {
            "name": "Timeseries Inversion Residuals",
            "file_list": _glob("*.vrt", subdir="timeseries_inversion_residuals"),
        },
        {
            "name": "Estimated Phase quality",
            "file_list": _glob("*.vrt", subdir="estimated_phase_quality"),
        },
        {
            "name": "SHP counts",
            "file_list": _glob("*.vrt", subdir="shp_counts"),
        },
        {
            "name": "Water Mask",
            "file_list": _glob("*.vrt", subdir="water_mask"),
        },
        {
            "name": "Unwrapper Mask",
            "file_list": _glob("*.vrt", subdir="unwrapper_mask"),
        },
        {
            "name": "Ionospheric Delay",
            "file_list": _glob("*vrt", subdir="corrections/ionospheric_delay"),
            "uses_spatial_ref": True,
            "algorithm": Algorithm.SHIFT.value,
        },
        {
            "name": "Perpendicular Baseline",
            "file_list": _glob("*vrt", subdir="corrections/perpendicular_baseline"),
        },
        {
            "name": "Solid Earth Tide",
            "file_list": _glob("*vrt", subdir="corrections/solid_earth_tide"),
            "uses_spatial_ref": True,
            "algorithm": Algorithm.SHIFT.value,
        },
    ]


def get_aligned_disp_s1_outputs(aligned_dir: Path | str):
    from glob import glob

    return [
        {
            "name": "Displacement",
            "file_list": sorted(glob(str(Path(aligned_dir) / "displacement*.tif"))),
            "uses_spatial_ref": True,
            "algorithm": Algorithm.SHIFT.value,
            "mask_file_list": sorted(
                glob(str(Path(aligned_dir) / "recommended_mask*.tif"))
            ),
        },
        {
            "name": "Short Wavelength Displacement",
            "file_list": sorted(
                glob(str(Path(aligned_dir) / "short_wavelength_displacement*.tif"))
            ),
        },
        {
            "name": "Connected Component Labels",
            "file_list": sorted(
                glob(str(Path(aligned_dir) / "connected_component_labels*.tif"))
            ),
        },
        {
            "name": "Re-wrapped phase",
            "file_list": sorted(glob(str(Path(aligned_dir) / "displacement*.tif"))),
            "algorithm": Algorithm.REWRAP.value,
        },
        {
            "name": "Persistent Scatterer Mask",
            "file_list": sorted(
                glob(str(Path(aligned_dir) / "persistent_scatterer_mask*.tif"))
            ),
        },
        {
            "name": "Temporal Coherence",
            "file_list": sorted(
                glob((str(Path(aligned_dir) / "temporal_coherence*.tif")))
            ),
        },
        {
            "name": "Phase Similarity",
            "file_list": sorted(
                glob((str(Path(aligned_dir) / "phase_similarity*.tif")))
            ),
        },
        {
            "name": "Timeseries Inversion Residuals",
            "file_list": sorted(
                glob(str(Path(aligned_dir) / "timeseries_inversion_residuals*.tif"))
            ),
        },
        {
            "name": "Estimated Phase quality",
            "file_list": sorted(
                glob(str(Path(aligned_dir) / "estimated_phase_quality*.tif"))
            ),
        },
        {
            "name": "SHP counts",
            "file_list": sorted(glob((str(Path(aligned_dir) / "shp_counts*.tif")))),
        },
        {
            "name": "Water Mask",
            "file_list": sorted(glob((str(Path(aligned_dir) / "water_mask.tif")))),
        },
    ]



import logging
from functools import partial
from pathlib import Path
from typing import Sequence

import h5py
from opera_utils import get_dates
from opera_utils.credentials import ASFCredentialEndpoints, AWSCredentials
from opera_utils.disp import open_h5
from osgeo import gdal
from tqdm.contrib.concurrent import process_map

from bowser.add_overviews import add_overviews

logger = logging.getLogger("bowser")


def process_netcdf_files(
    netcdf_files: Sequence[Path | str],
    output_dir: Path,
    datasets: list[str],
    max_workers: int = 5,
    strip_group_path: bool = False,
) -> None:
    """Process NetCDF files in the input directory, create VRT files, build overviews.

    Parameters
    ----------
    netcdf_files : Sequence[Path]
        Paths to input NetCDF files.
    output_dir : Path
        Path to the directory where output VRT files will be saved.
    datasets : list[str]
        list of dataset names to process from each NetCDF file.
    max_workers : int
        Number of parallel files to process.
        Default is 5.
    strip_group_path : bool
        If True, the output directory for the VRTs is only one level deep.
        Otherwise, uses the full HDF5 path as VRT path.
        Default is False.

    Returns
    -------
    None

    Notes
    -----
    This function processes all NetCDF files in the input directory, creates VRT files
    for each specified dataset, and builds overviews for the created VRT files.

    """
    # Ensure output directory exists
    out_path = Path(output_dir)
    out_path.mkdir(exist_ok=True, parents=True)

    if any(str(f).startswith("s3://") for f in netcdf_files):
        aws_credentials = AWSCredentials.from_asf(endpoint=ASFCredentialEndpoints.OPERA)
    else:
        aws_credentials = None

    func = partial(
        process_single_file,
        output_dir=output_dir,
        datasets=datasets,
        aws_credentials=aws_credentials,
        strip_group_path=strip_group_path,
    )

    process_map(
        func,
        netcdf_files,
        max_workers=max_workers,
    )


def process_single_file(
    netcdf_file: str,
    output_dir: Path,
    datasets: list[str],
    aws_credentials: AWSCredentials | None,
    strip_group_path: bool = False,
) -> None:
    """Create VRT files from subdatasets and build overviews.

    Parameters
    ----------
    netcdf_file : str
        Path to the input NetCDF file.
    output_dir : Path
        Path to the directory where output VRT files will be saved.
    datasets : list[str]
        list of dataset names to process from the NetCDF file.
    aws_credentials : AWSCredentials, optional
        Object containing temporary S3 credentials for remote datasets.
        Only usable on in-region EC2 instances.
    strip_group_path : bool
        If True, the output directory for the VRTs is only one level deep.
        Otherwise, uses the full HDF5 path as VRT path.
        Default is False.

    Returns
    -------
    None

    """
    # Extract date information from the filename

    try:
        fmt = "%Y%m%d"
        # TODO: NISAR ifgs may have 4 dates in them
        # Rethink how to make this filename here for multiple products
        dates = get_dates(netcdf_file, fmt=fmt)[:2]
        vrt_filename = f"{dates[0].strftime(fmt)}_{dates[1].strftime(fmt)}.vrt"
    except IndexError:
        # Date parsing failed: just use stem
        # TODO: NISAR holds ref/secondary as
        # /science/LSAR/identification/secondaryZeroDopplerEndTime
        vrt_filename = f"{str(netcdf_file).replace('/', '_')}.vrt"

    if str(netcdf_file).startswith("s3://"):
        hf = open_h5(netcdf_file, aws_credentials=aws_credentials)
        logger.debug(f"Read remote {netcdf_file}")
    else:
        hf = h5py.File(netcdf_file)

    for in_dataset in datasets:
        if in_dataset not in hf:
            continue
        out_dataset = in_dataset if not strip_group_path else in_dataset.split("/")[-1]
        cur_output_dir = output_dir / out_dataset
        cur_output_dir.mkdir(exist_ok=True, parents=True)

        vrt_path = cur_output_dir / vrt_filename

        # Create VRT file
        gdal.Translate(
            str(vrt_path),
            f"netcdf:{netcdf_file.replace('s3://', '/vsis3/')}:{in_dataset}",
            callback=gdal.TermProgress_nocb,
        )

        # Build overviews using GDAL function with compression
        add_overviews(vrt_path, external=True)



import os
import subprocess
import time
from pathlib import Path

from fastapi.testclient import TestClient

DATA_DIR = Path(__file__).parent / "data/geotiffs"


def test_setup_aligned_disp_s1_command(tmp_path):
    """Test that the setup-aligned-disp-s1 command creates a config file."""
    config_file = tmp_path / "bowser_rasters.json"

    # Run the setup command
    result = subprocess.run(
        ["bowser", "setup-aligned-disp-s1", str(DATA_DIR), "-o", config_file],
        capture_output=True,
        text=True,
        cwd=Path(__file__).parent.parent,
    )

    assert result.returncode == 0, f"Command failed: {result.stderr}"
    assert Path(config_file).exists()
    assert Path(config_file).stat().st_size > 0


def test_fastapi_app_startup(tmp_path):
    """Test that the FastAPI app can start with the sample data."""
    config_file = tmp_path / "bowser_rasters.json"
    # Run the setup command
    subprocess.run(
        ["bowser", "setup-aligned-disp-s1", str(DATA_DIR), "-o", config_file],
        capture_output=True,
        text=True,
        cwd=Path(__file__).parent.parent,
    )
    # Set up environment for test
    os.environ["BOWSER_DATASET_CONFIG_FILE"] = str(config_file)

    # Import after setting environment variables
    from bowser.main import app

    client = TestClient(app)

    # Test basic endpoints
    response = client.get("/mode")
    assert response.status_code == 200
    data = response.json()
    assert "mode" in data

    response = client.get("/datasets")
    assert response.status_code == 200


def test_bowser_cli_run_command(tmp_path):
    """Test that bowser run command doesn't crash immediately."""
    config_file = tmp_path / "bowser_rasters.json"
    # Run the setup command
    subprocess.run(
        ["bowser", "setup-aligned-disp-s1", str(DATA_DIR), "-o", config_file],
        capture_output=True,
        text=True,
        cwd=Path(__file__).parent.parent,
    )

    # Start the server in background and kill it quickly
    proc = subprocess.Popen(
        ["bowser", "run", "-f", str(config_file), "--port", "8999"],
        cwd=Path(__file__).parent.parent,
    )

    # Give it a moment to start up
    time.sleep(2)

    # Kill the process
    proc.terminate()
    proc.wait(timeout=5)

    # Check it didn't crash immediately
    assert proc.returncode != 1, "Server crashed on startup"



# import secrets
from typing import Union

from pydantic import AnyHttpUrl, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Settings for FastAPI app."""

    # Can be overridden by environment variable:
    BOWSER_DATASET_CONFIG_FILE: str = "bowser_rasters.json"
    BOWSER_STACK_DATA_FILE: str = ""
    LOG_LEVEL: str = "WARNING"

    # SECRET_KEY: str = secrets.token_urlsafe(32)
    # SERVER_NAME: str
    # SERVER_HOST: AnyHttpUrl
    # BACKEND_CORS_ORIGINS is a JSON-formatted list of origins
    # e.g: '["http://localhost", "http://localhost:4200", "http://localhost:3000", \
    # "http://localhost:8080", "http://local.dockertoolbox.tiangolo.com"]'
    BACKEND_CORS_ORIGINS: list[AnyHttpUrl] = []
    DOMAIN: str = "localhost"

    @classmethod
    @field_validator("BACKEND_CORS_ORIGINS", mode="before")
    def _assemble_cors_origins(cls, v: Union[str, list[str]]) -> Union[list[str], str]:
        if isinstance(v, str) and not v.startswith("["):
            return [i.strip() for i in v.split(",")]
        elif isinstance(v, (str, list)):
            return v
        raise ValueError(v)

    model_config = SettingsConfigDict(env_file=".env")


settings = Settings()



from __future__ import annotations

import os
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Any, Callable

import numpy as np
import simplejson
from pydantic import BaseModel, computed_field, field_validator
from rio_tiler.models import ImageData
from starlette import responses
from titiler.core.algorithm import BaseAlgorithm

from .readers import RasterStackReader
from .utils import list_bucket

# TODO: Prep COGS with
#  for f in `ls ig*[0-9].tif`; do
#    gdal_translate DERIVED_SUBDATASET:PHASE:$f phase_${f/.tif/.vrt};
#  done


class Algorithm(str, Enum):
    """Available algorithms."""

    PHASE = "phase"
    AMPLITUDE = "amplitude"
    SHIFT = "shift"
    REWRAP = "rewrap"


class Phase(BaseAlgorithm):
    """Creation algorithm for the phase of a complex raster."""

    def __call__(self, img: ImageData) -> ImageData:  # noqa: D102
        return _process_complex(img, np.angle)


class Amplitude(BaseAlgorithm):
    """Custom tile creation algorithm for amplitude of complex data."""

    def __call__(self, img: ImageData) -> ImageData:  # noqa: D102
        return _process_complex(img, np.abs)


class Rewrap(BaseAlgorithm):
    """Creation algorithm for re-wrapping unwrapped phase to (-pi, pi)."""

    scale_factor: float = 1.0

    def __call__(self, img: ImageData) -> ImageData:  # noqa: D102
        return ImageData(
            np.ma.mod(np.pi + (self.scale_factor * img.array), 2 * np.pi) - np.pi,
            # img. - self.shift,
            # np.ma.MaskedArray(data, mask=~mask),
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
        )


def _process_complex(
    img: ImageData, func: Callable[[np.ndarray], np.ndarray]
) -> ImageData:
    """Tile algorithm for derived statistics of a complex raster.

    See https://developmentseed.org/titiler/examples/code/tiler_with_custom_algorithm/
    """
    data = func(img.data)
    # Mask at 0  in addition to current mask
    mask = np.logical_or(np.isnan(data), data == 0)

    # Create output ImageData
    return ImageData(
        np.ma.MaskedArray(data, mask=mask),
        assets=img.assets,
        crs=img.crs,
        bounds=img.bounds,
    )


class Shift(BaseAlgorithm):
    """Apply a simple shift (to subtract a reference point."""

    # Parameters
    shift: float | None = 0
    nan_to_zero: bool = True

    def __call__(self, img: ImageData) -> ImageData:  # noqa: D102
        shift = np.nan_to_num(self.shift or 0) if self.nan_to_zero else self.shift
        return ImageData(
            img.array - shift,
            assets=img.assets,
            crs=img.crs,
            bounds=img.bounds,
        )


class RasterGroup(BaseModel):
    """A group of rasters to view."""

    name: str
    file_list: list[str | Path]
    mask_file_list: list[str | Path] = []
    nodata: float | None = None
    uses_spatial_ref: bool = False
    algorithm: str | None = None
    mask_min_value: float = 0.1
    file_date_fmt: str | None = "%Y%m%d"
    _reader: RasterStackReader

    @field_validator("file_list")
    @classmethod
    def _ensure_string(cls, v: Any):
        return list(map(os.fspath, v))

    def model_post_init(self, __context: Any) -> None:  # noqa: D102
        super().model_post_init(__context)

        self._reader = RasterStackReader.from_file_list(
            self.file_list,
            bands=1,
            keep_open=False,
            num_threads=3,
            nodata=self.nodata,
            file_date_fmt=self.file_date_fmt,
        )

    @computed_field
    def bounds(self) -> tuple[float, float, float, float]:
        """Project (left, bottom, right, top) bounds in the raster's CRS."""
        return self._reader.bounds

    @computed_field
    def latlon_bounds(self) -> tuple[float, float, float, float]:
        """Geographical (left, bottom, right, top) bounds of the raster group."""
        return self._reader.latlon_bounds

    # @functools.cached_property
    @computed_field
    def x_values(self) -> list[str | int]:
        """Vales to use for the x axis of a time series plot."""
        # if len(self.file_list) == 1:
        dates = self._reader.dates
        if not dates or any(d is None for d in dates):
            # otherwise, use indexes
            x_values = np.arange(len(self.file_list)).tolist()
        else:
            x_values = [_format_dates(*k) for k in dates]  # type: ignore[misc]

        return x_values

    @classmethod
    def from_glob(
        cls,
        glob_str: str,
        *,
        name: str,
        algorithm: str | None = None,
        nodata: float | None = None,
    ) -> RasterGroup:
        """Construct a RasterGroup from a glob string.

        Can be either a local path, or an s3 url.
        """
        file_list = _find_files(glob_str)
        if not file_list:
            raise ValueError("No files found.")

        return cls(
            name=name,
            file_list=_find_files(glob_str),
            algorithm=algorithm,
            nodata=nodata,
        )


def _find_files(glob_str: str) -> list[str]:
    if "*" not in glob_str:
        file_list = [glob_str]
    elif glob_str.startswith("s3://"):
        # Need to split the '*' from the rest of the path
        file_list = list_bucket(full_bucket_glob=glob_str)
    else:
        file_list = sorted(glob(glob_str))
    return file_list


def _format_dates(*dates, fmt="%Y%m%d") -> str:
    return "_".join(f"{d.strftime(fmt)}" for d in dates)


# https://github.com/developmentseed/titiler/blob/0fddd7ed268557e82a5e1520cdd7fdf084afa1b8/src/titiler/core/titiler/core/resources/responses.py#L15
class JSONResponse(responses.JSONResponse):
    """Custom JSON Response."""

    def render(self, content: Any) -> bytes:
        """Render JSON.

        Same defaults as starlette.responses.JSONResponse.render but allow NaN
        to be replaced by null using simplejson
        """
        return simplejson.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            ignore_nan=True,
        ).encode("utf-8")



import json
import subprocess
from functools import cache


class CredentialsError(Exception):
    """Raised when AWS credentials have expired."""


def list_bucket(
    bucket: str | None = None,
    prefix: str | None = None,
    suffix: str | None = None,
    full_bucket_glob: str | None = None,
    aws_profile: str | None = None,
    num_workers: int = 10,
) -> list[str]:
    """Use `s5cmd` to quickly list items in a bucket.

    Parameters
    ----------
    bucket : str
        Name of the bucket.
    prefix : str, optional
        Prefix to filter by, by default "".
    suffix : str, optional
        Suffix to filter by, by default "".
    full_bucket_glob : str, optional
        Alternate to prefix/suffix. Full glob to filter by, by default None.
    aws_profile : str, optional
        AWS profile to use, by default None.
    num_workers : int, default = 10
        Number of workers to use for parallel downloads.

    Returns
    -------
    list[str]
        List of items in the bucket.
    """
    cmd = ["s5cmd", "--json", "--numworkers", str(num_workers)]
    if aws_profile:
        cmd += ["--profile", aws_profile]

    if full_bucket_glob:
        bucket_str = full_bucket_glob
    else:
        bucket_str = f"s3://{bucket}"
        if prefix:
            bucket_str += f"/{prefix}"
        if suffix:
            bucket_str += f"*{suffix}"
    cmd += ["ls", bucket_str.strip()]
    try:
        p = subprocess.run(cmd, capture_output=True, text=True, check=True, timeout=90)
    except subprocess.CalledProcessError as e:
        if "ExpiredToken" in e.stderr:
            raise CredentialsError(
                "Error downloading files: AWS credentials have expired."
            ) from e
        raise RuntimeError(f"Error listing bucket {bucket_str}: {e.stderr}") from e

    out: list[str] = []
    for line in p.stdout.splitlines():
        item = json.loads(line)
        if item["type"] == "directory":
            continue
        out.append(item["key"])
    return out


@cache
def generate_colorbar(cmap_name: str) -> bytes:
    """Generate a colorbar image using a specified matplotlib colormap.

    Parameters
    ----------
    cmap_name : str
        Name of the colormap to be used.

    Returns
    -------
    bytes
        WEBP image data as bytes
    """
    from io import BytesIO

    import matplotlib.pyplot as plt
    import numpy as np

    gradient = np.array([[0, 1]])
    fig, ax = plt.subplots(figsize=(9, 1.5))
    ax.imshow(gradient, cmap=cmap_name)
    ax.set_visible(False)

    # Create colorbar in a tightly fitted axis
    cax = fig.add_axes([0.05, 0.2, 0.9, 0.6])
    colorbar = plt.colorbar(
        ax.imshow(gradient, cmap=cmap_name), cax=cax, orientation="horizontal"
    )
    colorbar.set_ticks([])

    # Adjust subplot parameters and apply tight layout
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0)

    buf = BytesIO()
    plt.savefig(buf, format="webp", bbox_inches="tight", pad_inches=0)
    plt.close(fig)
    buf.seek(0)
    return buf.read()


def desensitize_mpl_case():
    """Make `cmap` case insensitive by registering repeated cmap names."""
    import matplotlib as mpl

    cmap_names = list(mpl.colormaps)
    for name in cmap_names:
        if name.lower() in mpl.colormaps:
            continue
        mpl.colormaps.register(mpl.colormaps[name], name=name.lower())



import logging
from enum import Enum
from os import fspath
from pathlib import Path
from typing import Sequence

import click
from osgeo import gdal
from tqdm.contrib.concurrent import thread_map

PathOrStr = Path | str

gdal.UseExceptions()

logger = logging.getLogger("bowser")


class Resampling(Enum):
    """GDAL resampling algorithm."""

    NEAREST = "nearest"
    AVERAGE = "average"
    LANCZOS = "lanczos"


# TODO: This thing can now be done with the 3.11 GDAL cli
# https://gdal.org/en/stable/programs/gdal_raster_overview_add.html#gdal-raster-overview-add
# gdal raster overview add --config COMPRESS_OVERVIEW=LZW data.tif 2 4 8 16


def add_overviews(
    file_path: Path | str,
    overview_levels: Sequence[int] = [4, 8, 16, 32, 64],
    resampling: Resampling = Resampling.NEAREST,
    external: bool = False,
    compression: str = "LZW",
):
    """Add GDAL compressed overviews to an existing file.

    Parameters
    ----------
    file_path : Path
        Path to the file to process.
    overview_levels : list
        List of overview levels to add.
    resampling : str or Resampling
        GDAL resampling algorithm for overviews.
        Default = "nearest"
    external : bool, default = False
        Use external overviews (.ovr files).
    compression: str, default = "LZW"
        Compression algorithm to use for overviews.
        See https://gdal.org/programs/gdaladdo.html for options.
    """
    flags = gdal.GA_Update if not external else gdal.GA_ReadOnly
    ds = gdal.Open(fspath(file_path), flags)
    if ds.GetRasterBand(1).GetOverviewCount() > 0:
        logger.info("%s already has overviews. Skipping.")
        return
    raster = gdal.Open(fspath(file_path), flags)
    gdal.SetConfigOption("COMPRESS_OVERVIEW", compression)
    gdal.SetConfigOption("GDAL_NUM_THREADS", "2")
    raster.BuildOverviews(resampling.value, overview_levels)


def process_files(
    file_paths: Sequence[PathOrStr],
    levels: Sequence[int],
    resampling: Resampling,
    max_workers: int = 5,
):
    """Process files to add GDAL overviews and compression.

    Parameters
    ----------
    file_paths : Sequence[PathOrStr]
        Sequence of file paths to process.
    levels : Sequence[int]
        Sequence of overview levels to add.
    resampling : str or Resampling
        GDAL resampling algorithm for overviews.
        Default = "nearest"
    max_workers : int, default = 5
        Number of parallel threads to run.
    """
    thread_map(
        lambda file_path: add_overviews(
            Path(file_path),
            overview_levels=list(levels),
            resampling=resampling,
        ),
        file_paths,
        max_workers=max_workers,
    )


@click.command()
@click.argument("file_paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--levels",
    "-l",
    multiple=True,
    default=[4, 8, 16, 32, 64],
    type=int,
    help="Overview levels to add.",
    show_default=True,
)
@click.option(
    "--resampling",
    "-r",
    type=click.Choice([r.value for r in Resampling]),
    default="nearest",
    help="Resampling algorithm to use when building overviews",
    show_default=True,
)
@click.option(
    "--max-workers",
    "-n",
    default=5,
    type=int,
    help="Number of parallel files to process",
)
def addo(file_paths, levels, resampling, max_workers):
    """Add compressed GDAL overviews to files."""
    return process_files(
        file_paths=file_paths,
        levels=levels,
        resampling=Resampling(resampling),
        max_workers=max_workers,
    )
